// **************************************************************************
// * This file is part of the FreeFileSync project. It is distributed under *
// * GNU General Public License: http://www.gnu.org/licenses/gpl.html       *
// * Copyright (C) 2008-2010 ZenJu (zhnmju123 AT gmx.de)                    *
// **************************************************************************
//
#include "../resources.h"
#include "../../shared/fileHandling.h"
#include <wx/filefn.h>
#include <wx/file.h>
#include <wx/filename.h>
#include "unittest.h"
#include <wx/log.h>
#include "../../comparison.h"
#include "../../synchronization.h"
#include "../../algorithm.h"
#include "../../shared/globalFunctions.h"
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>
#include "../../shared/stringConv.h"

using namespace FreeFileSync;
using namespace boost::lambda;

unsigned int getTotalElementNoSub(const FreeFileSync::HierarchyObject& hierObj)
{
    unsigned int total = 0;
    std::for_each(hierObj.subDirs.begin(), hierObj.subDirs.end(), boost::ref(total) += bind(getTotalElementNoSub, _1));

    return hierObj.subDirs.size() + hierObj.subFiles.size() + total;
}

unsigned int getTotalElementNo(const FreeFileSync::FolderComparison& folderCmp)
{
    unsigned int total = 0;
    std::for_each(folderCmp.begin(), folderCmp.end(), boost::ref(total) += bind(getTotalElementNoSub, _1));
    return total;
}


std::vector<wxString> getSampleNames(const unsigned int count)
{
    std::vector<wxString> output;
    for (unsigned int i = 0; i < count; ++i)
    {
        wxString newName;
        newName += wxChar('A') + i%26;
        newName += wxChar('a') + (i*i)%25;
        newName += wxChar('A') + (i + 3)%24;
        newName += wxChar('A') + i%11 + 10;
        newName += wxT('語');
        if (i%2) newName+= wxChar('.');
        newName += wxChar('a') + (i + count + 8)%23;
        newName += wxChar('A') + (i + 13)%22;
        newName += wxChar('a') + (i + count + 15)%21;

        output.push_back(newName);
    }
    return output;
}


void foldersAreValidForComparison(const std::vector<FolderPairCfg>& folderPairs, StatusHandler* statusUpdater);

void syncByTimeDate(const wxString& leftDir, const wxString& rightDir, const SyncConfiguration& syncCfg, FolderComparison& output)
{
    output.clear();

    std::vector<FolderPairCfg> formattedFolders;
    FolderPairCfg newPair(
        FreeFileSync::getFormattedDirectoryName(wxToZ(leftDir)),
        FreeFileSync::getFormattedDirectoryName(wxToZ(rightDir)),
        BaseFilter::FilterRef(new NullFilter),
        syncCfg);
    formattedFolders.push_back(newPair);

    foldersAreValidForComparison(formattedFolders, NULL); //if not valid: dereferences NULL-pointer and aborts

    DummyStatusUpdater dummyStatus;

    //COMPARE DIRECTORIES
    xmlAccess::OptionalDialogs dummyWarnings;
    dummyWarnings.warningSignificantDifference = false;
    dummyWarnings.warningUnresolvedConflicts   = false;

    FreeFileSync::CompareProcess comparison(false, 0, false, dummyWarnings, &dummyStatus);
    comparison.startCompareProcess(formattedFolders,
                                   CMP_BY_TIME_SIZE,
                                   output);

    //START SYNCHRONIZATION
    FreeFileSync::SyncProcess synchronization(false, false, dummyWarnings, false, false, dummyStatus);

    std::vector<FolderPairSyncCfg> fpSyncCfg;
    fpSyncCfg.push_back(
        FolderPairSyncCfg(false,
                          FreeFileSync::DELETE_PERMANENTLY,
                          Zstring()));

    synchronization.startSynchronizationProcess(fpSyncCfg, output);

    //all rows should be processed (with the exception of conflicts
    SyncStatistics finalStat(output);
    ASSERT_EQ(finalStat.getCreate() + finalStat.getOverwrite() + finalStat.getDelete(), 0);
    ASSERT_TRUE(finalStat.getDataToProcess() == 0);

    //compare again to get status after sync
    output.clear();
    //create new CompareProcess instance because of buffering!
    FreeFileSync::CompareProcess comparison2(false, 0, false, dummyWarnings, &dummyStatus);
    comparison2.startCompareProcess(formattedFolders,
                                    CMP_BY_TIME_SIZE,
                                    output);
}


bool synchronizationNeeded(const SyncStatistics& statisticsTotal);


void createTestDirectories(const wxString& leftDir, const wxString& rightDir, std::vector<wxString>& directoriesRelative, std::vector<wxString>& filesRelative)
{
    //clean up
    ASSERT_NO_THROW(FreeFileSync::removeDirectory(wxToZ(leftDir)));
    ASSERT_NO_THROW(FreeFileSync::removeDirectory(wxToZ(rightDir)));

    ASSERT_FALSE(wxDirExists(leftDir));
    ASSERT_FALSE(wxDirExists(rightDir));

    //build vector of files and directories for testing
    directoriesRelative.clear();
    filesRelative.clear();

    directoriesRelative.push_back(wxEmptyString);
    unsigned int totalNumberOfObjects = 0;
    for (int k = 1; k <= 3; ++k) //directory hierarchy to create
    {
        std::vector<wxString> newNames = getSampleNames(10/k);
        totalNumberOfObjects += newNames.size() * directoriesRelative.size() * 2;

        std::vector<wxString> tmp;
        for (std::vector<wxString>::iterator i = directoriesRelative.begin(); i != directoriesRelative.end(); ++i)
        {
            for (std::vector<wxString>::iterator j = newNames.begin(); j != newNames.end(); ++j)
            {
                const wxString newDir = *i + zToWx(globalFunctions::FILE_NAME_SEPARATOR) + *j;
                tmp.push_back(newDir);

                const wxString newFile = *i + zToWx(globalFunctions::FILE_NAME_SEPARATOR) + wxT("file") + *j;
                filesRelative.push_back(newFile);
            }
        }
        tmp.swap(directoriesRelative);
    }

    ASSERT_NO_THROW(
    {
        //create directories
        wxLogNull dummy;
        for (std::vector<wxString>::iterator i = directoriesRelative.begin(); i != directoriesRelative.end(); ++i)
            FreeFileSync::createDirectory(wxToZ(leftDir + *i), Zstring(), false);
    });

    ASSERT_NO_THROW(FreeFileSync::createDirectory(wxToZ(rightDir), Zstring(), false));
    ASSERT_TRUE(wxDirExists(rightDir));

    //check that all directories on left have been created successfully
    ASSERT_TRUE(wxDirExists(leftDir));
    ASSERT_TRUE(wxDirExists(rightDir));
    for (std::vector<wxString>::iterator i = directoriesRelative.begin(); i != directoriesRelative.end(); ++i)
        ASSERT_TRUE(wxDirExists(leftDir + *i));

    //create files
    for (std::vector<wxString>::iterator i = filesRelative.begin(); i != filesRelative.end(); ++i)
    {
        wxFile newFile(leftDir + *i, wxFile::write);
        newFile.Write(wxT("dummy text"));
    }

    //check that all files have been created successfully
    for (std::vector<wxString>::iterator i = filesRelative.begin(); i != filesRelative.end(); ++i)
        ASSERT_TRUE(wxFileExists(leftDir + *i));

    //fill right directory through sync "left->right"
    //###########################
    FolderComparison currentGridData;
    SyncConfiguration syncConfig;
    syncConfig.automatic       = false;
    syncConfig.exLeftSideOnly  = FreeFileSync::SYNC_DIR_RIGHT;
    syncConfig.exRightSideOnly = FreeFileSync::SYNC_DIR_NONE;
    syncConfig.leftNewer       = FreeFileSync::SYNC_DIR_NONE;
    syncConfig.rightNewer      = FreeFileSync::SYNC_DIR_NONE;
    syncConfig.different       = FreeFileSync::SYNC_DIR_NONE;

    ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig, currentGridData));

    //check if both folders contain the same files

    //make sure the correct number of files and directories was created
    ASSERT_EQ(getTotalElementNo(currentGridData), totalNumberOfObjects);
    ASSERT_FALSE(synchronizationNeeded(SyncStatistics(currentGridData)));

    //check that all directories on right have been copied successfully
    for (std::vector<wxString>::iterator i = directoriesRelative.begin(); i != directoriesRelative.end(); ++i)
        ASSERT_TRUE(wxDirExists(rightDir + *i));

    //check that all files have been created successfully on right
    for (std::vector<wxString>::iterator i = filesRelative.begin(); i != filesRelative.end(); ++i)
        ASSERT_TRUE(wxFileExists(rightDir + *i));
}



void getDiffNumbersSub(const HierarchyObject& hierObj, int& equal, int& leftOnly, int& rightOnly, int& leftNewer, int& rightNewer, int& different, int& conflict)
{
    for (HierarchyObject::SubFileMapping::const_iterator i = hierObj.subFiles.begin(); i != hierObj.subFiles.end(); ++i)
        switch (i->getCategory())
        {
        case FILE_LEFT_SIDE_ONLY:
            leftOnly++;
            break;
        case FILE_RIGHT_SIDE_ONLY:
            rightOnly++;
            break;
        case FILE_RIGHT_NEWER:
            rightNewer++;
            break;
        case FILE_LEFT_NEWER:
            leftNewer++;
            break;
        case FILE_DIFFERENT:
            different++;
            break;
        case FILE_EQUAL:
            equal++;
            break;
        case FILE_CONFLICT:
            conflict++;
            break;
        }

    for (HierarchyObject::SubDirMapping::const_iterator i = hierObj.subDirs.begin(); i != hierObj.subDirs.end(); ++i)
    {
        //recursion
        getDiffNumbersSub(*i, equal, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);

        switch (i->getCategory())
        {
        case FILE_LEFT_SIDE_ONLY:
            leftOnly++;
            break;
        case FILE_RIGHT_SIDE_ONLY:
            rightOnly++;
            break;
        case FILE_EQUAL:
            equal++;
            break;

        case FILE_RIGHT_NEWER:
        case FILE_LEFT_NEWER:
        case FILE_DIFFERENT:
        case FILE_CONFLICT:
        default:
            FAIL();
            break;
        }
    }
}


void getDiffNumbers(const FolderComparison& folderCmp, int& leftOnly, int& rightOnly, int& leftNewer, int& rightNewer, int& different, int& conflict)
{
    int equal  = 0;
    leftOnly   = 0;
    rightOnly  = 0;
    leftNewer  = 0;
    rightNewer = 0;
    different  = 0;
    conflict   = 0;

    std::for_each(folderCmp.begin(), folderCmp.end(),
                  bind(getDiffNumbersSub, _1,
                       boost::ref(equal),
                       boost::ref(leftOnly),
                       boost::ref(rightOnly),
                       boost::ref(leftNewer),
                       boost::ref(rightNewer),
                       boost::ref(different),
                       boost::ref(conflict)));

    ASSERT_EQ(int(getTotalElementNo(folderCmp)), leftOnly + rightOnly + leftNewer + rightNewer + different + equal + conflict);
}


//predicate-formatter for asserting that two dates are equal
testing::AssertionResult assertEqual(const char* a_expr, const char* b_expr, const wxDateTime& a, const wxDateTime& b)
{
    if (a == b)
        return testing::AssertionSuccess();

    testing::Message msg;
    msg << a_expr << " = " << a.Format().c_str() <<
    " and " << b_expr << " = " << b.Format().c_str() << " not equal!";

    return testing::AssertionFailure(msg);
}


//predicate-formatter for asserting that two long integers are equal
testing::AssertionResult assertEqual(const char* a_expr, const char* b_expr, const wxULongLong& a, const wxULongLong& b)
{
    if (a == b)
        return testing::AssertionSuccess();

    testing::Message msg;
    msg << a_expr << " = " << a.ToString().c_str() <<
    " and " << b_expr << " = " << b.ToString().c_str() << " not equal!";

    return testing::AssertionFailure(msg);
}


//predicate-formatter for asserting that two wxStrings are equal
testing::AssertionResult assertEqual(const char* a_expr, const char* b_expr, const wxString& a, const wxString& b)
{
    if (a == b)
        return testing::AssertionSuccess();

    testing::Message msg;
    msg << a_expr << " = " << a.c_str() <<
    " and " << b_expr << " = " << b.c_str() << " not equal!";

    return testing::AssertionFailure(msg);
}


class SynchronizationTest : public testing::Test
{
protected:
    // Per-test-case set-up.
    // Called before the first test in this test case.
    static void SetUpTestCase()
    {
        leftDir  = wxT("\\unit_test_left");
        rightDir = wxT("\\unit_test_right");

        //create and fill directories on HD for testing
        ASSERT_NO_THROW(createTestDirectories(leftDir, rightDir, directoriesRel, filesRel));

        //create differences of all types in folder structure
        restoreFolderDiff();

        //retrieve comparison result through dummy sync
        FolderComparison cmpResult2;
        SyncConfiguration syncConfig2;
        syncConfig2.automatic       = false;
        syncConfig2.exLeftSideOnly  = FreeFileSync::SYNC_DIR_NONE;
        syncConfig2.exRightSideOnly = FreeFileSync::SYNC_DIR_NONE;
        syncConfig2.leftNewer       = FreeFileSync::SYNC_DIR_NONE;
        syncConfig2.rightNewer      = FreeFileSync::SYNC_DIR_NONE;
        syncConfig2.different       = FreeFileSync::SYNC_DIR_NONE;

        ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig2, cmpResult2));

        //check that "compare" returns the right numbers of each type
        initialLineCount = getTotalElementNo(cmpResult2);
        getDiffNumbers(cmpResult2, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
        ASSERT_EQ(leftOnly, 2);
        ASSERT_EQ(rightOnly, 2);
        ASSERT_EQ(leftNewer, 1);
        ASSERT_EQ(rightNewer, 1);
        ASSERT_EQ(different, 0);
        ASSERT_EQ(conflict, 1);
    }

    // Per-test-case tear-down.
    // Called after the last test in this test case.
    static void TearDownTestCase()
    {
        //clean up
        ASSERT_NO_THROW(FreeFileSync::removeDirectory(wxToZ(leftDir)));
        ASSERT_NO_THROW(FreeFileSync::removeDirectory(wxToZ(rightDir)));

        ASSERT_FALSE(wxDirExists(leftDir));
        ASSERT_FALSE(wxDirExists(rightDir));
    }


    static void restoreFolderDiff() //create differences in folder structure of all types
    {
        //left side only
        if (wxFileExists(rightDir + filesRel[20]))
            ASSERT_TRUE(wxRemoveFile(rightDir + filesRel[20]));
        if (wxDirExists(rightDir + directoriesRel[20]))
            ASSERT_NO_THROW(FreeFileSync::removeDirectory(wxToZ(rightDir + directoriesRel[20])));
        if (!wxFileExists(leftDir + filesRel[20])) //restore test files/folders
        {
            wxFile newFile(leftDir + filesRel[20], wxFile::write);
            newFile.Write(wxT("adskfjödsa"));
        }
        if (!wxDirExists(leftDir + directoriesRel[20]))
            ASSERT_NO_THROW(FreeFileSync::createDirectory(wxToZ(leftDir + directoriesRel[20]), Zstring(), false));

        //right side only
        if (wxFileExists(leftDir + filesRel[10]))
            ASSERT_TRUE(wxRemoveFile(leftDir + filesRel[10]));
        if (wxDirExists(leftDir + directoriesRel[10]))
            ASSERT_NO_THROW(FreeFileSync::removeDirectory(wxToZ(leftDir + directoriesRel[10])));

        if (!wxFileExists(rightDir + filesRel[10])) //restore test files/folders
        {
            wxFile newFile(rightDir + filesRel[10], wxFile::write);
            newFile.Write(wxT("adskfjödsa"));
        }
        if (!wxDirExists(rightDir + directoriesRel[10]))
            ASSERT_NO_THROW(FreeFileSync::createDirectory(wxToZ(rightDir + directoriesRel[10]), Zstring(), false));

        //left newer
        {
            wxDateTime modTime;
            wxFileName file(leftDir + filesRel[30]);
            ASSERT_TRUE(file.GetTimes(NULL, &modTime, NULL)); //get modification time
            modTime.Add(wxDateSpan(0,0,0,1));
            ASSERT_TRUE(file.SetTimes(NULL, &modTime, NULL)); //get modification time
        }

        //right newer
        {
            wxDateTime modTime;
            wxFileName file(rightDir + filesRel[40]);
            ASSERT_TRUE(file.GetTimes(NULL, &modTime, NULL)); //get modification time
            modTime.Add(wxDateSpan(0,0,0,1));
            ASSERT_TRUE(file.SetTimes(NULL, &modTime, NULL)); //get modification time
        }

        //conflict
        {
            wxDateTime modTime;
            wxFileName file(rightDir + filesRel[50]);
            ASSERT_TRUE(file.GetTimes(NULL, &modTime, NULL)); //get modification time
            {
                wxFile newFile(rightDir + filesRel[50], wxFile::write_append);
                newFile.Write(wxT("asfsafdsafdsafadsfdsaf"));
            }

            ASSERT_TRUE(file.SetTimes(NULL, &modTime, NULL)); //get modification time
        }
    }

//shared resources
    static unsigned int initialLineCount; //nr of rows in compareResult-vector
    static wxString leftDir;
    static wxString rightDir;

    static std::vector<wxString> directoriesRel; //relative names
    static std::vector<wxString> filesRel;

//temporal variables
    static SyncConfiguration syncConfig;
    static int leftOnly;
    static int rightOnly;
    static int leftNewer;
    static int rightNewer;
    static int different;
    static int conflict;
    static wxDateTime modTimeLeft;
    static wxDateTime modTimeRight;
    static wxULongLong fileSizeLeft;
    static wxULongLong fileSizeRight;
};

unsigned int SynchronizationTest::initialLineCount;
wxString SynchronizationTest::leftDir;
wxString SynchronizationTest::rightDir;
std::vector<wxString> SynchronizationTest::directoriesRel;
std::vector<wxString> SynchronizationTest::filesRel;
SyncConfiguration SynchronizationTest::syncConfig;
int SynchronizationTest::leftOnly;
int SynchronizationTest::rightOnly;
int SynchronizationTest::leftNewer;
int SynchronizationTest::rightNewer;
int SynchronizationTest::different;
int SynchronizationTest::conflict;
wxDateTime SynchronizationTest::modTimeLeft;
wxDateTime SynchronizationTest::modTimeRight;
wxULongLong SynchronizationTest::fileSizeLeft;
wxULongLong SynchronizationTest::fileSizeRight;


TEST_F(SynchronizationTest, LeftSideOnly)
{
    //left side only(1)
    //--------------------------------------------------------
    syncConfig.automatic       = false;
    syncConfig.exLeftSideOnly  = FreeFileSync::SYNC_DIR_RIGHT;
    syncConfig.exRightSideOnly = FreeFileSync::SYNC_DIR_NONE;
    syncConfig.leftNewer       = FreeFileSync::SYNC_DIR_NONE;
    syncConfig.rightNewer      = FreeFileSync::SYNC_DIR_NONE;
    syncConfig.different       = FreeFileSync::SYNC_DIR_NONE;

    FolderComparison cmpResult;
    ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig, cmpResult));

    ASSERT_TRUE(wxFileExists(rightDir + filesRel[20]));  //new files/dirs created
    ASSERT_TRUE(wxDirExists(rightDir + directoriesRel[20]));

    getDiffNumbers(cmpResult, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
    ASSERT_EQ(leftOnly, 0);
    ASSERT_EQ(rightOnly, 2);
    ASSERT_EQ(leftNewer, 1);
    ASSERT_EQ(rightNewer, 1);
    ASSERT_EQ(different, 0);
    ASSERT_EQ(conflict, 1);

    ASSERT_EQ(getTotalElementNo(cmpResult), initialLineCount);

    //restore test files/folders
    ASSERT_TRUE(wxRemoveFile(rightDir + filesRel[20]));
    ASSERT_NO_THROW(FreeFileSync::removeDirectory(wxToZ(rightDir + directoriesRel[20])));


    //left side only(2)
    //--------------------------------------------------------
    syncConfig.automatic       = false;
    syncConfig.exLeftSideOnly  = FreeFileSync::SYNC_DIR_LEFT;
    syncConfig.exRightSideOnly = FreeFileSync::SYNC_DIR_NONE;
    syncConfig.leftNewer       = FreeFileSync::SYNC_DIR_NONE;
    syncConfig.rightNewer      = FreeFileSync::SYNC_DIR_NONE;
    syncConfig.different       = FreeFileSync::SYNC_DIR_NONE;

    ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig, cmpResult));

    ASSERT_FALSE(wxFileExists(leftDir + filesRel[20]));  //files/dirs deleted
    ASSERT_FALSE(wxDirExists(leftDir + directoriesRel[20]));

    getDiffNumbers(cmpResult, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
    ASSERT_EQ(leftOnly, 0);
    ASSERT_EQ(rightOnly, 2);
    ASSERT_EQ(leftNewer, 1);
    ASSERT_EQ(rightNewer, 1);
    ASSERT_EQ(different, 0);
    ASSERT_EQ(conflict, 1);

    ASSERT_EQ(getTotalElementNo(cmpResult), initialLineCount - 2);

    //restore test files/folders
    ASSERT_FALSE(wxFileExists(leftDir + filesRel[20]));
    {
        wxFile newFile(leftDir + filesRel[20], wxFile::write);
        newFile.Write(wxT("adskfjödsa"));
    }
    ASSERT_NO_THROW(FreeFileSync::createDirectory(wxToZ(leftDir + directoriesRel[20]), Zstring(), false));
}


TEST_F(SynchronizationTest, RightSideOnly)
{
    //right side only(1)
    //--------------------------------------------------------
    syncConfig.automatic       = false;
    syncConfig.exLeftSideOnly  = FreeFileSync::SYNC_DIR_NONE;
    syncConfig.exRightSideOnly = FreeFileSync::SYNC_DIR_RIGHT;
    syncConfig.leftNewer       = FreeFileSync::SYNC_DIR_NONE;
    syncConfig.rightNewer      = FreeFileSync::SYNC_DIR_NONE;
    syncConfig.different       = FreeFileSync::SYNC_DIR_NONE;

    FolderComparison cmpResult;
    ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig, cmpResult));

    ASSERT_FALSE(wxFileExists(rightDir + filesRel[10]));  //files/dirs deleted
    ASSERT_FALSE(wxDirExists(rightDir + directoriesRel[10]));

    getDiffNumbers(cmpResult, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
    ASSERT_EQ(leftOnly, 2);
    ASSERT_EQ(rightOnly, 0);
    ASSERT_EQ(leftNewer, 1);
    ASSERT_EQ(rightNewer, 1);
    ASSERT_EQ(different, 0);
    ASSERT_EQ(conflict, 1);

    ASSERT_EQ(getTotalElementNo(cmpResult), initialLineCount - 2); //one file and directory were deleted

    //restore test files/folders
    ASSERT_FALSE(wxFileExists(rightDir + filesRel[10]));
    {
        wxFile newFile(rightDir + filesRel[10], wxFile::write);
        newFile.Write(wxT("adskfjödsa"));
    }
    ASSERT_NO_THROW(FreeFileSync::createDirectory(wxToZ(rightDir + directoriesRel[10]), Zstring(), false));


    //right side only(2)
    //--------------------------------------------------------
    syncConfig.automatic       = false;
    syncConfig.exLeftSideOnly  = FreeFileSync::SYNC_DIR_NONE;
    syncConfig.exRightSideOnly = FreeFileSync::SYNC_DIR_LEFT;
    syncConfig.leftNewer       = FreeFileSync::SYNC_DIR_NONE;
    syncConfig.rightNewer      = FreeFileSync::SYNC_DIR_NONE;
    syncConfig.different       = FreeFileSync::SYNC_DIR_NONE;

    ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig, cmpResult));

    ASSERT_TRUE(wxFileExists(leftDir + filesRel[10]));  //files/dirs created
    ASSERT_TRUE(wxDirExists(leftDir + directoriesRel[10]));

    getDiffNumbers(cmpResult, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
    ASSERT_EQ(leftOnly, 2);
    ASSERT_EQ(rightOnly, 0);
    ASSERT_EQ(leftNewer, 1);
    ASSERT_EQ(rightNewer, 1);
    ASSERT_EQ(different, 0);
    ASSERT_EQ(conflict, 1);

    ASSERT_EQ(getTotalElementNo(cmpResult), initialLineCount);

    //restore test files/folders
    ASSERT_TRUE(wxRemoveFile(leftDir + filesRel[10]));
    ASSERT_NO_THROW(FreeFileSync::removeDirectory(wxToZ(leftDir + directoriesRel[10])));
}


TEST_F(SynchronizationTest, LeftNewer)
{
    //left newer(1)
    //--------------------------------------------------------
    syncConfig.automatic       = false;
    syncConfig.exLeftSideOnly  = FreeFileSync::SYNC_DIR_NONE;
    syncConfig.exRightSideOnly = FreeFileSync::SYNC_DIR_NONE;
    syncConfig.leftNewer       = FreeFileSync::SYNC_DIR_RIGHT;
    syncConfig.rightNewer      = FreeFileSync::SYNC_DIR_NONE;
    syncConfig.different       = FreeFileSync::SYNC_DIR_NONE;

    modTimeLeft = wxFileName(leftDir + filesRel[30]).GetModificationTime();
    modTimeRight = wxFileName(rightDir + filesRel[30]).GetModificationTime();
    ASSERT_FALSE(modTimeLeft == modTimeRight);

    FolderComparison cmpResult;
    ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig, cmpResult));

    modTimeRight = wxFileName(rightDir + filesRel[30]).GetModificationTime();
    ASSERT_PRED_FORMAT2(assertEqual, modTimeLeft, modTimeRight); //right file should be overwritten => mod time from left file should match

    getDiffNumbers(cmpResult, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
    ASSERT_EQ(leftOnly, 2);
    ASSERT_EQ(rightOnly, 2);
    ASSERT_EQ(leftNewer, 0);
    ASSERT_EQ(rightNewer, 1);
    ASSERT_EQ(different, 0);
    ASSERT_EQ(conflict, 1);

    ASSERT_EQ(getTotalElementNo(cmpResult), initialLineCount);

    //restore test files/folders
    {
        wxDateTime modTime;
        wxFileName file(leftDir + filesRel[30]);
        ASSERT_TRUE(file.GetTimes(NULL, &modTime, NULL)); //get modification time
        modTime.Add(wxDateSpan(0,0,0,1));
        ASSERT_TRUE(file.SetTimes(NULL, &modTime, NULL)); //get modification time
    }


    //left newer(2)
    //--------------------------------------------------------
    syncConfig.automatic       = false;
    syncConfig.exLeftSideOnly  = FreeFileSync::SYNC_DIR_NONE;
    syncConfig.exRightSideOnly = FreeFileSync::SYNC_DIR_NONE;
    syncConfig.leftNewer       = FreeFileSync::SYNC_DIR_LEFT;
    syncConfig.rightNewer      = FreeFileSync::SYNC_DIR_NONE;
    syncConfig.different       = FreeFileSync::SYNC_DIR_NONE;

    modTimeLeft = wxFileName(leftDir + filesRel[30]).GetModificationTime();
    modTimeRight = wxFileName(rightDir + filesRel[30]).GetModificationTime();
    ASSERT_FALSE(modTimeLeft == modTimeRight);

    ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig, cmpResult));

    modTimeLeft = wxFileName(leftDir + filesRel[30]).GetModificationTime();
    ASSERT_PRED_FORMAT2(assertEqual, modTimeLeft, modTimeRight); //left file should be overwritten => mod time from right file should match

    getDiffNumbers(cmpResult, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
    ASSERT_EQ(leftOnly, 2);
    ASSERT_EQ(rightOnly, 2);
    ASSERT_EQ(leftNewer, 0);
    ASSERT_EQ(rightNewer, 1);
    ASSERT_EQ(different, 0);
    ASSERT_EQ(conflict, 1);

    ASSERT_EQ(getTotalElementNo(cmpResult), initialLineCount);

    //restore test files/folders
    {
        wxDateTime modTime;
        wxFileName file(leftDir + filesRel[30]);
        ASSERT_TRUE(file.GetTimes(NULL, &modTime, NULL)); //get modification time
        modTime.Add(wxDateSpan(0,0,0,1));
        ASSERT_TRUE(file.SetTimes(NULL, &modTime, NULL)); //get modification time
    }
}


TEST_F(SynchronizationTest, RightNewer)
{
    //right newer(1)
    //--------------------------------------------------------
    syncConfig.automatic       = false;
    syncConfig.exLeftSideOnly  = FreeFileSync::SYNC_DIR_NONE;
    syncConfig.exRightSideOnly = FreeFileSync::SYNC_DIR_NONE;
    syncConfig.leftNewer       = FreeFileSync::SYNC_DIR_NONE;
    syncConfig.rightNewer      = FreeFileSync::SYNC_DIR_RIGHT;
    syncConfig.different       = FreeFileSync::SYNC_DIR_NONE;

    modTimeLeft = wxFileName(leftDir + filesRel[40]).GetModificationTime();
    modTimeRight = wxFileName(rightDir + filesRel[40]).GetModificationTime();
    ASSERT_FALSE(modTimeLeft == modTimeRight);

    FolderComparison cmpResult;
    ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig, cmpResult));

    modTimeRight = wxFileName(rightDir + filesRel[40]).GetModificationTime();
    ASSERT_PRED_FORMAT2(assertEqual, modTimeLeft, modTimeRight); //right file should be overwritten => mod time from left file should match

    getDiffNumbers(cmpResult, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
    ASSERT_EQ(leftOnly, 2);
    ASSERT_EQ(rightOnly, 2);
    ASSERT_EQ(leftNewer, 1);
    ASSERT_EQ(rightNewer, 0);
    ASSERT_EQ(different, 0);
    ASSERT_EQ(conflict, 1);

    ASSERT_EQ(getTotalElementNo(cmpResult), initialLineCount);
    //restore test files/folders
    {
        wxDateTime modTime;
        wxFileName file(rightDir + filesRel[40]);
        ASSERT_TRUE(file.GetTimes(NULL, &modTime, NULL)); //get modification time
        modTime.Add(wxDateSpan(0,0,0,1));
        ASSERT_TRUE(file.SetTimes(NULL, &modTime, NULL)); //get modification time
    }


    //right newer(2)
    //--------------------------------------------------------
    syncConfig.automatic       = false;
    syncConfig.exLeftSideOnly  = FreeFileSync::SYNC_DIR_NONE;
    syncConfig.exRightSideOnly = FreeFileSync::SYNC_DIR_NONE;
    syncConfig.leftNewer       = FreeFileSync::SYNC_DIR_NONE;
    syncConfig.rightNewer      = FreeFileSync::SYNC_DIR_LEFT;
    syncConfig.different       = FreeFileSync::SYNC_DIR_NONE;

    modTimeLeft = wxFileName(leftDir + filesRel[40]).GetModificationTime();
    modTimeRight = wxFileName(rightDir + filesRel[40]).GetModificationTime();
    ASSERT_FALSE(modTimeLeft == modTimeRight);

    ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig, cmpResult));

    modTimeLeft = wxFileName(leftDir + filesRel[40]).GetModificationTime();
    ASSERT_PRED_FORMAT2(assertEqual, modTimeLeft, modTimeRight); //left file should be overwritten => mod time from right file should match

    getDiffNumbers(cmpResult, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
    ASSERT_EQ(leftOnly, 2);
    ASSERT_EQ(rightOnly, 2);
    ASSERT_EQ(leftNewer, 1);
    ASSERT_EQ(rightNewer, 0);
    ASSERT_EQ(different, 0);
    ASSERT_EQ(conflict, 1);

    ASSERT_EQ(getTotalElementNo(cmpResult), initialLineCount);
    //restore test files/folders
    {
        wxDateTime modTime;
        wxFileName file(rightDir + filesRel[40]);
        ASSERT_TRUE(file.GetTimes(NULL, &modTime, NULL)); //get modification time
        modTime.Add(wxDateSpan(0,0,0,1));
        ASSERT_TRUE(file.SetTimes(NULL, &modTime, NULL)); //get modification time
    }
}


TEST_F(SynchronizationTest, OneWay)
{
    //one way sync(1)
    //--------------------------------------------------------
    syncConfig.automatic       = false;
    syncConfig.exLeftSideOnly  = FreeFileSync::SYNC_DIR_RIGHT;
    syncConfig.exRightSideOnly = FreeFileSync::SYNC_DIR_RIGHT;
    syncConfig.leftNewer       = FreeFileSync::SYNC_DIR_RIGHT;
    syncConfig.rightNewer      = FreeFileSync::SYNC_DIR_RIGHT;
    syncConfig.different       = FreeFileSync::SYNC_DIR_RIGHT;

    FolderComparison cmpResult;
    ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig, cmpResult));

    getDiffNumbers(cmpResult, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
    ASSERT_EQ(leftOnly, 0);
    ASSERT_EQ(rightOnly, 0);
    ASSERT_EQ(leftNewer, 0);
    ASSERT_EQ(rightNewer, 0);
    ASSERT_EQ(different, 0);
    ASSERT_EQ(conflict, 1);

    ASSERT_EQ(getTotalElementNo(cmpResult), initialLineCount - 2);
    ASSERT_FALSE(wxFileExists(rightDir + filesRel[10]));
    ASSERT_FALSE(wxDirExists(rightDir + directoriesRel[10]));
    //restore test files/folders
    restoreFolderDiff(); //a little bit less precise as specific file manipulation


    //one way sync(2)
    //--------------------------------------------------------
    syncConfig.automatic       = false;
    syncConfig.exLeftSideOnly  = FreeFileSync::SYNC_DIR_LEFT;
    syncConfig.exRightSideOnly = FreeFileSync::SYNC_DIR_LEFT;
    syncConfig.leftNewer       = FreeFileSync::SYNC_DIR_LEFT;
    syncConfig.rightNewer      = FreeFileSync::SYNC_DIR_LEFT;
    syncConfig.different       = FreeFileSync::SYNC_DIR_LEFT;

    ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig, cmpResult));

    getDiffNumbers(cmpResult, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
    ASSERT_EQ(leftOnly, 0);
    ASSERT_EQ(rightOnly, 0);
    ASSERT_EQ(leftNewer, 0);
    ASSERT_EQ(rightNewer, 0);
    ASSERT_EQ(different, 0);
    ASSERT_EQ(conflict, 1);

    ASSERT_EQ(getTotalElementNo(cmpResult), initialLineCount - 2);
    ASSERT_FALSE(wxFileExists(leftDir + filesRel[20]));
    ASSERT_FALSE(wxDirExists(leftDir + directoriesRel[20]));
    //restore test files/folders
    restoreFolderDiff(); //a little bit less precise as specific file manipulation
}


TEST_F(SynchronizationTest, TwoWay)
{
    //two way sync(1)
    //--------------------------------------------------------
    syncConfig.automatic       = false;
    syncConfig.exLeftSideOnly  = FreeFileSync::SYNC_DIR_RIGHT;
    syncConfig.exRightSideOnly = FreeFileSync::SYNC_DIR_LEFT;
    syncConfig.leftNewer       = FreeFileSync::SYNC_DIR_RIGHT;
    syncConfig.rightNewer      = FreeFileSync::SYNC_DIR_LEFT;
    syncConfig.different       = FreeFileSync::SYNC_DIR_LEFT; //should be irrelevant

    FolderComparison cmpResult;
    ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig, cmpResult));

    getDiffNumbers(cmpResult, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
    ASSERT_EQ(leftOnly, 0);
    ASSERT_EQ(rightOnly, 0);
    ASSERT_EQ(leftNewer, 0);
    ASSERT_EQ(rightNewer, 0);
    ASSERT_EQ(different, 0);
    ASSERT_EQ(conflict, 1);

    ASSERT_EQ(getTotalElementNo(cmpResult), initialLineCount);
    //restore test files/folders
    restoreFolderDiff(); //a little bit less precise as specific file manipulation


    //two way sync(2)
    //--------------------------------------------------------
    syncConfig.automatic       = false;
    syncConfig.exLeftSideOnly  = FreeFileSync::SYNC_DIR_LEFT;
    syncConfig.exRightSideOnly = FreeFileSync::SYNC_DIR_RIGHT;
    syncConfig.leftNewer       = FreeFileSync::SYNC_DIR_LEFT;
    syncConfig.rightNewer      = FreeFileSync::SYNC_DIR_RIGHT;
    syncConfig.different       = FreeFileSync::SYNC_DIR_LEFT; //should be irrelevant

    ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig, cmpResult));

    getDiffNumbers(cmpResult, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
    ASSERT_EQ(leftOnly, 0);
    ASSERT_EQ(rightOnly, 0);
    ASSERT_EQ(leftNewer, 0);
    ASSERT_EQ(rightNewer, 0);
    ASSERT_EQ(different, 0);
    ASSERT_EQ(conflict, 1);

    ASSERT_EQ(getTotalElementNo(cmpResult), initialLineCount - 4);
    ASSERT_FALSE(wxFileExists(rightDir + filesRel[10]));
    ASSERT_FALSE(wxDirExists(rightDir + directoriesRel[10]));
    ASSERT_FALSE(wxFileExists(leftDir + filesRel[20]));
    ASSERT_FALSE(wxDirExists(leftDir + directoriesRel[20]));
    //restore test files/folders
    restoreFolderDiff(); //a little bit less precise as specific file manipulation
}


TEST_F(SynchronizationTest, CompareByContent)
{
    const wxString newFileRel = wxT("\\abcdefg.dat");

    ASSERT_FALSE(wxFileExists(rightDir + newFileRel));
    ASSERT_FALSE(wxFileExists(leftDir + newFileRel));

    //create additional file pair with different contents
    {
        wxFile newFile(rightDir + newFileRel, wxFile::write);
        newFile.Write(wxT("123456789"));
    }
    {
        wxFile newFile(leftDir + newFileRel, wxFile::write);
        newFile.Write(wxT("123456799"));
    }

    //setup sample files

    std::vector<FolderPairCfg> formattedFolders;
    FolderPairCfg newPair(
        FreeFileSync::getFormattedDirectoryName(wxToZ(leftDir)),
        FreeFileSync::getFormattedDirectoryName(wxToZ(rightDir)),
        BaseFilter::FilterRef(new NullFilter),
        SyncConfiguration());
    formattedFolders.push_back(newPair);

    foldersAreValidForComparison(formattedFolders, NULL);

    DummyStatusUpdater dummyStatus;

    //COMPARE DIRECTORIES (by content)
    xmlAccess::OptionalDialogs dummyWarnings;
    dummyWarnings.warningSyncDatabase = false;

    FolderComparison cmpResult;
    FreeFileSync::CompareProcess comparison(false, 0, false, dummyWarnings, &dummyStatus);
    comparison.startCompareProcess(formattedFolders,
                                   CMP_BY_CONTENT,
                                   cmpResult);


    getDiffNumbers(cmpResult, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
    ASSERT_EQ(leftOnly, 2);
    ASSERT_EQ(rightOnly, 2);
    ASSERT_EQ(leftNewer, 0);
    ASSERT_EQ(rightNewer, 0);
    ASSERT_EQ(different, 2);
    ASSERT_EQ(conflict, 0);

    ASSERT_TRUE(wxRemoveFile(rightDir + newFileRel));
    ASSERT_TRUE(wxRemoveFile(leftDir + newFileRel));
}

